@@ -25,6 +25,8 @@ test_that("xml_child() errors if search is not numeric or character", {
 
 test_that("xml_length", {
   expect_equal(xml_length(x), 2)
+  all <- xml_find_all(x, "//*")
+  expect_equal(xml_length(all), c(2, 1, 0, 0))
 })
 
 test_that("xml_parent", {
@@ -36,3 +38,15 @@ test_that("xml_parents", {
     xml_name(xml_parents(xml_find_first(x, "//boo"))),
     c("bar", "foo"))
 })
+
+test_that("xml_root", {
+  doc <- xml_new_document()
+
+  expect_is(xml_root(doc), "xml_missing")
+
+  a <- xml_add_child(doc, "a")
+  b <- xml_add_child(doc, "b")
+
+  expect_that(xml_name(xml_root(b)), equals("a"))
+  expect_that(xml_name(xml_root(doc)), equals("a"))
+})
